#include <ModbusMaster.h>

/*
  When using a MAX485-compatible RS485 Transceiver.
  Rx/Tx is connected up to the hardware serial port, this will vary for different board types.
  The Data Enable (DE) and Receiver Enable (RE) pins are connected to the pins below:
  In some cases, the DE and RE pins can be connected and controlled via one pin.

  The two type of RS485 Transceivers which were used in this application are listed below:
    MAX485 TTL to RS-485 Interface Module
    RS422 / RS485 Shield for Arduino UNO

*/
#define MAX485_DE_RE 6
//#define MAX485_DE 6
//#define MAX485_RE 7
/*
  Arduino class library for communicating with Modbus slaves over RS232/485 (via RTU protocol).
*/
ModbusMaster modbus;

void preTransmission() {
  /* 
    This sets the DE/RE to high in order to enable data transmit and disable receive.
  */
  digitalWrite(MAX485_DE_RE, 1);
  //digitalWrite(MAX485_DE,1);
  //digitalWrite(MAX485_RE,1);
}

void postTransmission() {
  /* 
    This sets the DE/RE to low in order to disable data transmit and enable receive.
  */
  digitalWrite(MAX485_DE_RE, 0);
  //digitalWrite(MAX485_DE, 0);
  //digitalWrite(MAX485_RE, 0);
}

void PIDSupplyEnable() {
  uint8_t state;
  /*
    Set Transmit Buffer is used to format the data ready for transmission, .writeMultipleRegisters() will transmit the buffer in this application. 
    In buffer setting value '1' to buffer index position '0', this is due to the value '1' being required to enable the PID supply on the SDK.
    If more data would like to be sent, then .setTransmitBuffer() will have to be called for each position of the registers.
  */
  modbus.setTransmitBuffer(0, 1);

  /*
    Write Multiple Registers requires the function .setTransmitBuffer() to be populated.
    Writing to register 0x1248, value '1' referring to quantity of registeres to write to.
    This will enable the PID Supply for the SDK, visually can be confirmed by the LED on hardware.
  */
  state = modbus.writeMultipleRegisters(0x1248, 1);
  if (state == modbus.ku8MBSuccess) {
    Serial.println("Powering On PID Supply!");
  } else {
    Serial.println("Unable to Enable PID Supply!");}
}

void PIDSupplyDisable() {
  uint8_t state;
  /*
    Set Transmit Buffer is used to format the data ready for transmission, .writeMultipleRegisters() will transmit the buffer in this application. 
    In buffer setting value '0' to buffer index position '0', this is due to the value '0' being required to disable the PID supply on the SDK.
    If more data would like to be sent, then .setTransmitBuffer() will have to be called for each position of the registers.
  */
  modbus.setTransmitBuffer(0, 0);

  /*
    Write Multiple Registers requires the function .setTransmitBuffer() to be populated.
    Writing to register 0x1248, value '1' referring to quantity of registeres to write to.
    This will enable the PID Supply for the SDK, visually can be confirmed by the LED on hardware.
  */
  state = modbus.writeMultipleRegisters(0x1248, 1);
  if (state == modbus.ku8MBSuccess) {
    Serial.println("Powering Off PID Supply!");
  } else{
      Serial.println("Unable to Disable PID Supply!");
}
}

void setup() {
  pinMode(MAX485_DE_RE, OUTPUT);
  digitalWrite(MAX485_DE_RE, 0);


  /*
    Serial.begin() will need to reviewed based on hardware used.

    Refer to the following link for the different serial port configurations available.
    https://www.arduino.cc/reference/en/language/functions/communication/serial/

    Find some listed examples below:

    UNO R3. Rx = pin 0, Tx = pin 1 - Serial.begin()
    UNO R4. Rx = pin 0, Tx = pin 1 - Serial1.begin()
    MKR boards. Rx = pin 13, Tx = pin 14 - Serial1.begin()
    Nano (classic). Rx = pin 0, Tx = pin 1 - Serial.begin()

    Serial baudrate for this application can either be 9600 or 19200 due to SDK configuration availability.

    Please refer to the SDK datasheet in order to decide the suitable serial parity configuration.
      SERIAL_8E1 = Parity Even, 1 Stop Bit
      SERIAL_8N2 = Parity None, 2 Stop Bit  
      SERIAL_8O1 = Parity Odd,  1 Stop Bit

  */
  Serial1.begin(9600, SERIAL_8E1);

  /*
    Begins Modbus RTU data stream communication by selecting the slave address, followed by the communication method.
    Modbus Slave ID = 0x01
    Communication = Serial1 - UNO R4. Rx = pin 0, Tx = pin 1
  */
  modbus.begin(0x01, Serial1);

  /* 
    Serial.begin() in this application is being used for serial monitoring/terminal.
    If hardware requires use of Serial.begin() for Modbus communication then alternative serial library might be needed such as "SoftwareSerial.h".
  */
  Serial.begin(115200);

  /*
   If DE/RE pins are being used, allows us to configure the RS485 transceiver correctly.
  */
  modbus.preTransmission(preTransmission);
  modbus.postTransmission(postTransmission);

  /* 
    Delay added to
  */
  PIDSupplyDisable();
  delay(1000);
  PIDSupplyEnable();
  delay(1000);
  PIDSupplyDisable();
  delay(1000);
  PIDSupplyEnable();
  delay(1000);
}

union {
  uint16_t ui16[2];  // 16-bit Unsigned Integer
  float f32;         // 32-bit Float
  uint32_t ui32;     // 32-bit Unsigned Integer
  int32_t i32;       // 32-bit Signed Integer
} data;

uint8_t result;

void loop() {
  /* 
    Reading Input Registers 0x1260 - ADCMV.
    Slave ID = 0x01
    Function Code = 0x04
    Start Register = 0x12C0
    Quantity of Registers = 12 - Each register is 2 bytes(16 bits)
  */
  result = modbus.readInputRegisters(0x1260, 12);
  if (result == modbus.ku8MBSuccess) {
    data.ui16[0] = modbus.getResponseBuffer(0x00);
    data.ui16[1] = modbus.getResponseBuffer(0x01);
    Serial.print("RTD - raw ADC input in uV: ");
    Serial.println(data.ui32);
    data.ui16[0] = modbus.getResponseBuffer(0x02);
    data.ui16[1] = modbus.getResponseBuffer(0x03);
    Serial.print("10V - raw ADC input in uV: ");
    Serial.println(data.ui32);
    data.ui16[0] = modbus.getResponseBuffer(0x04);
    data.ui16[1] = modbus.getResponseBuffer(0x05);
    Serial.print("PID - raw ADC input in uV: ");
    Serial.println(data.ui32);
    data.ui16[0] = modbus.getResponseBuffer(0x06);
    data.ui16[1] = modbus.getResponseBuffer(0x07);
    Serial.print("LIGHT - raw ADC input in uV: ");
    Serial.println(data.ui32);
    data.ui16[0] = modbus.getResponseBuffer(0x08);
    data.ui16[1] = modbus.getResponseBuffer(0x09);
    Serial.print("TEMPERATURE - raw ADC input in uV: ");
    Serial.println(data.ui32);
    data.ui16[0] = modbus.getResponseBuffer(0x0A);
    data.ui16[1] = modbus.getResponseBuffer(0x0B);
    Serial.print("4-20 - raw ADC input in ADC units: ");
    Serial.println(data.ui32);
  } else {
    Serial.println("Unable to read data!");}
      delay(2000);
  /* 
    Reading Input Registers 0x12C0 - SIGNAL.
    Slave ID = 0x01
    Function Code = 0x04
    Start Register = 0x12C0
    Quantity of Registers = 12 - Each register is 2 bytes(16 bits)
  */
  result = modbus.readInputRegisters(0x12C0, 12);
  if (result == modbus.ku8MBSuccess) {
    data.ui16[0] = modbus.getResponseBuffer(0x00);
    data.ui16[1] = modbus.getResponseBuffer(0x01);
    Serial.print("RTD - Temperature in degC or deg: ");
    Serial.println(data.f32);
    data.ui16[0] = modbus.getResponseBuffer(0x02);
    data.ui16[1] = modbus.getResponseBuffer(0x03);
    Serial.print("10V - Scaled Input: ");
    Serial.println(data.f32);
    data.ui16[0] = modbus.getResponseBuffer(0x04);
    data.ui16[1] = modbus.getResponseBuffer(0x05);
    Serial.print("PID - Signal in ppm: ");
    Serial.println(data.f32);
    data.ui16[0] = modbus.getResponseBuffer(0x06);
    data.ui16[1] = modbus.getResponseBuffer(0x07);
    Serial.print("Lamp - state: ");
    Serial.println(data.f32);
    data.ui16[0] = modbus.getResponseBuffer(0x08);
    data.ui16[1] = modbus.getResponseBuffer(0x09);
    Serial.print("TEMPERATURE - Internal Temperature in degC: ");
    Serial.println(data.f32);
    data.ui16[0] = modbus.getResponseBuffer(0x0A);
    data.ui16[1] = modbus.getResponseBuffer(0x0B);
    Serial.print("4-20 - Scaled Input: ");
    Serial.println(data.f32);
  } else {
    Serial.println("Unable to read data!");
  }
  delay(2000);
  /* 
    Reading Holding Registers 0x1200 - RTC.
    Slave ID = 0x01
    Function Code = 0x03
    Start Register = 0x12C0
    Quantity of Registers = 16 - Each register is 2 bytes(16 bits)
  */
  result = modbus.readHoldingRegisters(0x1200, 16);
  if (result == modbus.ku8MBSuccess) {
    data.ui16[0] = modbus.getResponseBuffer(0x00);
    data.ui16[1] = modbus.getResponseBuffer(0x01);
    Serial.print("Year: ");
    Serial.println(data.ui32);
    data.ui16[0] = modbus.getResponseBuffer(0x02);
    data.ui16[1] = modbus.getResponseBuffer(0x03);
    Serial.print("Month: ");
    Serial.println(data.ui32);
    data.ui16[0] = modbus.getResponseBuffer(0x04);
    data.ui16[1] = modbus.getResponseBuffer(0x05);
    Serial.print("Day: ");
    Serial.println(data.ui32);
    data.ui16[0] = modbus.getResponseBuffer(0x06);
    data.ui16[1] = modbus.getResponseBuffer(0x07);
    Serial.print("Hour: ");
    Serial.println(data.ui32);
    data.ui16[0] = modbus.getResponseBuffer(0x08);
    data.ui16[1] = modbus.getResponseBuffer(0x09);
    Serial.print("Minute: ");
    Serial.println(data.ui32);
    data.ui16[0] = modbus.getResponseBuffer(0x0A);
    data.ui16[1] = modbus.getResponseBuffer(0x0B);
    Serial.print("Second: ");
    Serial.println(data.ui32);
    data.ui16[0] = modbus.getResponseBuffer(0x0C);
    data.ui16[1] = modbus.getResponseBuffer(0x0D);
    Serial.print("Day of Week: ");
    Serial.println(data.ui32);
    data.ui16[0] = modbus.getResponseBuffer(0x0E);
    data.ui16[1] = modbus.getResponseBuffer(0x0F);
    Serial.print("Day of Year: ");
    Serial.println(data.ui32);
  } else {
    Serial.println("Unable to read data!");
  }
  delay(2000);
}

//Check the Max485 TTL
